"""barcode step."""

import os
import re
import subprocess
import sys
import glob
from collections import defaultdict, Counter
from itertools import combinations, product

import pandas as pd
import pysam
from xopen import xopen

import celescope.tools.utils as utils
from celescope.tools.__init__ import __PATTERN_DICT__
from celescope.tools.Chemistry import Chemistry
from celescope.tools.Step import Step, s_common


MIN_T = 10

def seq_ranges(seq, pattern_dict):
    # get subseq with intervals in arr and concatenate
    return ''.join([seq[x[0]:x[1]]for x in pattern_dict])


def get_seq_list(seq, pattern_dict, abbr):
    # get subseq list
    return (seq[item[0]: item[1]] for item in pattern_dict[abbr])


@utils.add_log
def parse_pattern(pattern):
    # 解析接头结构，返回接头结构字典
    # key: 字母表示的接头, value: 碱基区间列表
    # eg.: C8L10C8L10C8U8T30
    # defaultdict(<type 'list'>:
    # {'C': [[0, 8], [18, 26], [36, 44]], 'U': [[44, 52]], 'L': [[8, 18], [26, 36]], 'T': [[52, 82]]})
    pattern_dict = defaultdict(list)
    p = re.compile(r'([CLUNT])(\d+)')
    tmp = p.findall(pattern)
    if not tmp:
        parse_pattern.logger.error(f'Invalid pattern: {pattern}')
        sys.exit()
    start = 0
    for item in tmp:
        end = start + int(item[1])
        pattern_dict[item[0]].append([start, end])
        start = end
    return pattern_dict


def get_scope_bc(chemistry):
    import celescope
    root_path = os.path.dirname(celescope.__file__)
    if chemistry == 'scopeV1':
        return None, None
    linker_f = glob.glob(f'{root_path}/data/chemistry/{chemistry}/linker*')[0]
    whitelist_f = f'{root_path}/data/chemistry/{chemistry}/bclist'
    return linker_f, whitelist_f


def read_fastq(f):
    """
    Deprecated!
    Return tuples: (name, sequence, qualities).
    qualities is a string and it contains the unmodified, encoded qualities.
    """
    i = 3
    for i, line in enumerate(f):
        if i % 4 == 0:
            assert line.startswith('@'), ("Line {0} in FASTQ file is expected to start with '@', "
                                          "but found {1!r}".format(i + 1, line[:10]))
            name = line.strip()[1:]
        elif i % 4 == 1:
            sequence = line.strip()
        elif i % 4 == 2:
            line = line.strip()
            assert line.startswith('+'), ("Line {0} in FASTQ file is expected to start with '+', "
                                          "but found {1!r}".format(i + 1, line[:10]))
        elif i % 4 == 3:
            qualities = line.rstrip('\n\r')
            yield name, sequence, qualities
    if i % 4 != 3:
        raise Exception("FASTQ file ended prematurely")


def ord2chr(q, offset=33):
    return chr(int(q) + offset)


def qual_int(char, offset=33):
    return ord(char) - offset


def low_qual(quals, minQ, num):
    # print(ord('/')-33)           14
    return True if len([q for q in quals if qual_int(q) < minQ]) > num else False


def check_seq(seq_file, pattern_dict, seq_abbr):
    length = 0
    for item in pattern_dict[seq_abbr]:
        start = item[0]
        end = item[1]
        length += end - start
    with open(seq_file, 'r') as fh:
        for seq in fh:
            seq = seq.strip()
            if seq == '':
                continue
            if len(seq) != length:
                raise Exception(
                    f'length of L in pattern ({length}) do not equal to length in {seq_file} ({len(seq)}) !')


def get_mismatch(seq, n_mismatch=1, bases='ACGTN'):
    '''
    choose locations where there's going to be a mismatch using combinations
    and then construct all satisfying lists using product

    Return:
    all mismatch <= n_mismatch dict. Key: mismatch_seq, value: orig_seq
    '''
    seq_set = set()
    mismatch_dict = {}
    seq_len = len(seq)
    if n_mismatch > seq_len:
        n_mismatch = seq_len
    for locs in combinations(range(seq_len), n_mismatch):
        seq_locs = [[base] for base in seq]
        for loc in locs:
            seq_locs[loc] = list(bases)
        for poss in product(*seq_locs):
            seq_set.add(''.join(poss))
    for mismatch_seq in seq_set:
        mismatch_dict[mismatch_seq] = seq
    return mismatch_dict


@utils.add_log
def get_all_mismatch(seq_list, n_mismatch=1):
    '''
    Return:
    all mismatch <= n_mismatch dict. Key: mismatch_seq, value: orig_seq from seq_file
    '''
    mismatch_dict = {}
    correct_set = set()

    for seq in seq_list:
        seq = seq.strip()
        if seq == '':
            continue
        correct_set.add(seq)
        mismatch_dict.update(get_mismatch(seq, n_mismatch=n_mismatch))

    return correct_set, mismatch_dict


def check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list):
    '''
    Return bool_valid, bool_corrected
    '''
    bool_valid = True
    bool_corrected = False
    corrected_seq = ''
    for index, seq in enumerate(seq_list):
        if seq not in correct_set_list[index]:
            if seq not in mismatch_dict_list[index]:
                bool_valid = False
                return bool_valid, bool_corrected, corrected_seq
            else:
                bool_corrected = True
                corrected_seq += mismatch_dict_list[index][seq]
        else:
            corrected_seq += seq
    return bool_valid, bool_corrected, corrected_seq


class Barcode(Step):

    '''barcode step class
    '''   
    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)

        self.fq1_list = args.fq1.split(",")
        self.fq2_list = args.fq2.split(",")
        self.fq_number = len(self.fq1_list)
        if self.fq_number != len(self.fq2_list):
            raise Exception('fastq1 and fastq2 do not have same file number!')
        if args.chemistry == 'auto':
            ch = Chemistry(args.fq1)
            self.chemistry_list = ch.check_chemistry()
        else:
            self.chemistry_list = [args.chemistry] * self.fq_number
        self.barcode_corrected_num = 0
        self.linker_corrected_num = 0
        self.total_num = 0
        self.clean_num = 0
        self.no_polyT_num = 0
        self.lowQual_num = 0
        self.no_linker_num = 0
        self.no_barcode_num = 0
        self.barcode_qual_Counter = Counter()
        self.umi_qual_Counter = Counter()
        if args.gzip:
            suffix = ".gz"
        else:
            suffix = ""
        self.out_fq2 = f'{self.outdir}/{self.sample}_2.fq{suffix}'
        self.nopolyT = args.nopolyT
        self.noLinker = args.noLinker
        self.bool_probe = False
        if args.probe_file and args.probe_file != 'None':
            self.bool_probe = True
            self.probe_count_dic = utils.genDict(dim=3)
            self.valid_count_dic = utils.genDict(dim=2)
            self.probe_dic = utils.read_fasta(args.probe_file)
            self.reads_without_probe = 0
        self.pattern = args.pattern
        self.linker = args.linker
        self.whitelist = args.whitelist
        self.lowNum = args.lowNum
        self.lowQual = args.lowQual
        self.allowNoPolyT = args.allowNoPolyT
        self.allowNoLinker = args.allowNoLinker   

    @utils.add_log
    def run(self):

        fh3 = xopen(self.out_fq2, 'w')

        if self.nopolyT:
            fh1_without_polyT = xopen(self.outdir + '/noPolyT_1.fq', 'w')
            fh2_without_polyT = xopen(self.outdir + '/noPolyT_2.fq', 'w')

        if self.noLinker:
            fh1_without_linker = xopen(self.outdir + '/noLinker_1.fq', 'w')
            fh2_without_linker = xopen(self.outdir + '/noLinker_2.fq', 'w')

        for i in range(self.fq_number):

            chemistry = self.chemistry_list[i]
            lowNum = int(self.lowNum)
            Barcode.run.logger.info(f'lowQual score: {self.lowQual}')
            lowQual = int(self.lowQual)
            if chemistry == 'scopeV1':
                lowNum = min(0, lowNum)
                lowQual = max(10, lowQual)
                Barcode.run.logger.info(f'scopeV1: lowNum={lowNum}, lowQual={lowQual} ')
            # get linker and whitelist
            bc_pattern = __PATTERN_DICT__[chemistry]
            if (bc_pattern):
                (linker, whitelist) = get_scope_bc(chemistry)
            else:
                bc_pattern = self.pattern
                linker = self.linker
                whitelist = self.whitelist
            if not bc_pattern:
                raise Exception("invalid bc_pattern!")

            # parse pattern to dict, C8L10C8L10C8U8
            # defaultdict(<type 'list'>, {'C': [[0, 8], [18, 26], [36, 44]], 'U':
            # [[44, 52]], 'L': [[8, 18], [26, 36]]})
            pattern_dict = parse_pattern(bc_pattern)

            bool_T = True if 'T' in pattern_dict else False
            bool_L = True if 'L' in pattern_dict else False
            bool_whitelist = (whitelist is not None) and whitelist != "None"
            C_len = sum([item[1] - item[0] for item in pattern_dict['C']])

            if bool_whitelist:
                seq_list, _ = utils.read_one_col(whitelist)
                barcode_correct_set, barcode_mismatch_dict = get_all_mismatch(seq_list, n_mismatch=1)
                barcode_correct_set_list = [barcode_correct_set] * 3
                barcode_mismatch_dict_list = [barcode_mismatch_dict] * 3
            if bool_L:
                seq_list, _ = utils.read_one_col(linker)
                check_seq(linker, pattern_dict, "L")
                linker_correct_set_list = []
                linker_mismatch_dict_list = []
                start = 0
                for item in pattern_dict['L']:
                    end = start + item[1] - item[0]
                    linker_seq_list = [seq[start:end] for seq in seq_list]
                    linker_correct_set, linker_mismatch_dict = get_all_mismatch(linker_seq_list, n_mismatch=2)
                    linker_correct_set_list.append(linker_correct_set)
                    linker_mismatch_dict_list.append(linker_mismatch_dict)
                    start = end

            fq1 = pysam.FastxFile(self.fq1_list[i], persist=False)
            fq2 = pysam.FastxFile(self.fq2_list[i], persist=False)

            for entry1 in fq1:
                entry2 = next(fq2)
                header1, seq1, qual1 = entry1.name, entry1.sequence, entry1.quality
                header2, seq2, qual2 = entry2.name, entry2.sequence, entry2.quality
                self.total_num += 1

                # polyT filter
                if bool_T and (not self.allowNoPolyT):
                    polyT = seq_ranges(seq1, pattern_dict['T'])
                    if polyT.count('T') < MIN_T:
                        self.no_polyT_num += 1
                        if self.nopolyT:
                            fh1_without_polyT.write(
                                '@%s\n%s\n+\n%s\n' % (header1, seq1, qual1))
                            fh2_without_polyT.write(
                                '@%s\n%s\n+\n%s\n' % (header2, seq2, qual2))
                        continue

                # lowQual filter                
                C_U_quals_ascii = seq_ranges(
                    qual1, pattern_dict['C'] + pattern_dict['U'])
                # C_U_quals_ord = [ord(q) - 33 for q in C_U_quals_ascii]
                if lowQual > 0 and low_qual(C_U_quals_ascii, lowQual, lowNum):
                    self.lowQual_num += 1
                    continue

                # linker filter
                if bool_L and (not self.allowNoLinker):
                    seq_list = get_seq_list(seq1, pattern_dict, 'L')
                    bool_valid, bool_corrected, _ = check_seq_mismatch(
                        seq_list, linker_correct_set_list, linker_mismatch_dict_list)
                    if not bool_valid:
                        self.no_linker_num += 1
                        if self.noLinker:
                            fh1_without_linker.write(
                                '@%s\n%s\n+\n%s\n' % (header1, seq1, qual1))
                            fh2_without_linker.write(
                                '@%s\n%s\n+\n%s\n' % (header2, seq2, qual2))
                        continue
                    elif bool_corrected:
                        self.linker_corrected_num += 1
                
                # barcode filter
                seq_list = get_seq_list(seq1, pattern_dict, 'C')
                if bool_whitelist:
                    bool_valid, bool_corrected, corrected_seq = check_seq_mismatch(
                        seq_list, barcode_correct_set_list, barcode_mismatch_dict_list)

                    if not bool_valid:
                        self.no_barcode_num += 1
                        continue
                    elif bool_corrected:
                        self.barcode_corrected_num += 1
                    cb = corrected_seq
                else:
                    cb = "".join(seq_list)

                umi = seq_ranges(seq1, pattern_dict['U'])

                self.clean_num += 1

                if self.bool_probe:
                    # valid count
                    read_name_probe = 'None'
                    self.valid_count_dic[cb][umi] += 1

                    # output probe UMi and read count
                    find_probe = False
                    for probe_name in self.probe_dic:
                        probe_seq = self.probe_dic[probe_name]
                        probe_seq = probe_seq.upper()
                        if seq1.find(probe_seq) != -1:
                            self.probe_count_dic[probe_name][cb][umi] += 1
                            read_name_probe = probe_name
                            find_probe = True
                            break

                    if not find_probe:
                        self.reads_without_probe += 1

                self.barcode_qual_Counter.update(C_U_quals_ascii[:C_len])
                self.umi_qual_Counter.update(C_U_quals_ascii[C_len:])

                fh3.write(f'@{cb}_{umi}_{self.total_num}\n{seq2}\n+\n{qual2}\n')
            Barcode.run.logger.info(self.fq1_list[i] + ' finished.')
        fh3.close()

        # logging
        Barcode.run.logger.info(
            f'processed reads: {utils.format_number(self.total_num)}. '
            f'valid reads: {utils.format_number(self.clean_num)}. '
        )

        Barcode.run.logger.info(f'no polyT reads number : {self.no_polyT_num}')
        Barcode.run.logger.info(f'low qual reads number: {self.lowQual_num}')
        Barcode.run.logger.info(f'no_linker: {self.no_linker_num}')
        Barcode.run.logger.info(f'no_barcode: {self.no_barcode_num}')
        Barcode.run.logger.info(f'corrected linker: {self.linker_corrected_num}')
        Barcode.run.logger.info(f'corrected barcode: {self.barcode_corrected_num}')

        if self.clean_num == 0:
            raise Exception(
                'no valid reads found! please check the --chemistry parameter.')

        if self.bool_probe:
            # total probe summary
            total_umi = 0
            total_valid_read = 0
            for cb in self.valid_count_dic:
                total_umi += len(self.valid_count_dic[cb])
                total_valid_read += sum(self.valid_count_dic[cb].values())

            # probe summary
            count_list = []
            for probe_name in self.probe_dic:
                UMI_count = 0
                read_count = 0
                if probe_name in self.probe_count_dic:
                    for cb in self.probe_count_dic[probe_name]:
                        UMI_count += len(self.probe_count_dic[probe_name][cb])
                        read_count += sum(self.probe_count_dic[probe_name][cb].values())
                count_list.append(
                    {"probe_name": probe_name, "UMI_count": UMI_count, "read_count": read_count})

            df_count = pd.DataFrame(count_list, columns=[
                                    "probe_name", "read_count", "UMI_count"])

            def format_percent(x):
                x = str(round(x*100, 2))+"%"
                return x
            df_count["read_fraction"] = (
                df_count["read_count"]/total_valid_read).apply(format_percent)
            df_count["UMI_fraction"] = (
                df_count["UMI_count"]/total_umi).apply(format_percent)
            df_count.sort_values(by="UMI_count", inplace=True, ascending=False)
            df_count_file = self.outdir + '/' + self.sample + '_probe_count.tsv'
            df_count.to_csv(df_count_file, sep="\t", index=False)

        # stat
        BarcodesQ30 = sum([self.barcode_qual_Counter[k] for k in self.barcode_qual_Counter if k >= ord2chr(
            30)]) / float(sum(self.barcode_qual_Counter.values())) * 100
        UMIsQ30 = sum([self.umi_qual_Counter[k] for k in self.umi_qual_Counter if k >= ord2chr(
            30)]) / float(sum(self.umi_qual_Counter.values())) * 100

        def cal_percent(x): return "{:.2%}".format((x + 0.0) / self.total_num)
        stat_info = '''
            Raw Reads: %s
            Valid Reads: %s(%s)
            Q30 of Barcodes: %.2f%%
            Q30 of UMIs: %.2f%%
        '''
        with open(self.outdir + '/stat.txt', 'w') as fh:
            stat_info = stat_info % (utils.format_number(self.total_num), utils.format_number(self.clean_num),
                                    cal_percent(self.clean_num), BarcodesQ30,
                                    UMIsQ30)
            stat_info = re.sub(r'^\s+', r'', stat_info, flags=re.M)
            fh.write(stat_info)
        
        # self.fastqc()
        self.clean_up()


    @utils.add_log
    def fastqc(self):
        cmd = ['fastqc', '-t', str(self.thread), '-o', self.outdir, self.out_fq2]
        Barcode.fastqc.logger.info('%s' % (' '.join(cmd)))
        subprocess.check_call(cmd)


@utils.add_log
def barcode(args):
    step_name = "barcode"
    barcode_obj = Barcode(args, step_name)
    barcode_obj.run()


def get_opts_barcode(parser, sub_program=True):
    parser.add_argument('--pattern', help='')
    parser.add_argument('--whitelist', help='')
    parser.add_argument('--linker', help='')
    parser.add_argument('--lowQual', type=int,
                        help='max phred of base as lowQual, default=0', default=0)
    parser.add_argument(
        '--lowNum', type=int, help='max number with lowQual allowed, default=2', default=2)
    parser.add_argument('--nopolyT', action='store_true',
                        help='output nopolyT fq')
    parser.add_argument('--noLinker', action='store_true',
                        help='output noLinker fq')
    parser.add_argument('--probe_file', help="probe fasta file")
    parser.add_argument('--allowNoPolyT', help="allow reads without polyT", action='store_true')
    parser.add_argument('--allowNoLinker', help="allow reads without correct linker", action='store_true')
    parser.add_argument('--gzip', help="output gzipped fastq", action='store_true')
    parser.add_argument(
        '--chemistry', choices=__PATTERN_DICT__.keys(), help='chemistry version', default='auto')
    if sub_program:
        parser.add_argument('--fq1', help='read1 fq file', required=True)
        parser.add_argument('--fq2', help='read2 fq file', required=True)
        parser = s_common(parser)

    return parser



import os
import sys
import re
import subprocess
import logging
import pandas as pd
import pysam

from itertools import islice
from celescope.tools.utils import *
from celescope.tools.Step import Step


ADAPT = ['polyT=A{18}', 'p5=AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC']


def format_stat(cutadapt_log, sample_name):
    fh = open(cutadapt_log, 'r')
    stat_file = os.path.dirname(cutadapt_log) + '/stat.txt'
    # Total reads processed:...Total written (filtered):
    content = islice(fh, 9, 16)
    p_list = []
    for line in content:
        if line.strip() == '':
            continue
        line = re.sub(r'\s{2,}', r'', line)
        line = re.sub(r' bp', r'', line)
        line = re.sub(r'(?<=\d)\s+\(', r'(', line)
        line = line.strip()
        attr = line.split(":")
        p_list.append({"item": attr[0], "value": attr[1]})
    p_df = pd.DataFrame(p_list)
    p_df.iloc[0, 0] = 'Reads with Adapters'
    p_df.iloc[1, 0] = 'Reads too Short'
    p_df.iloc[2, 0] = 'Reads Written'
    p_df.iloc[3, 0] = 'Base Pairs Processed'
    p_df.iloc[4, 0] = 'Base Pairs Quality-Trimmed'
    p_df.iloc[5, 0] = 'Base Pairs Written'
    p_df.to_csv(stat_file, sep=':', index=False, header=None)

    fh.close()


@add_log
def read_adapter_fasta(adapter_fasta):
    '''
    return ['adapter1=AAA','adapter2=BBB']
    '''
    adapter_args = []
    if adapter_fasta and adapter_fasta != 'None':
        with pysam.FastxFile(adapter_fasta) as fh:
            for read in fh:
                adapter_args.append(f'{read.name}={read.sequence}')
    return adapter_args


@add_log
def cutadapt(args):

    step_name = "cutadapt"
    step = Step(args, step_name)

    adapter_args = read_adapter_fasta(args.adapter_fasta)
    adapter_args += ADAPT

    # run cutadapt
    adapt = []
    for a in adapter_args:
        adapt.append('-a')
        adapt.append(a)

    if args.gzip:
        suffix = ".gz"
    else:
        suffix = ""
    out_fq2 = f'{args.outdir}/{args.sample}_clean_2.fq{suffix}'
    cmd = ['cutadapt'] + adapt + ['-n',
                                  str(len(adapter_args)),
                                  '-j',
                                  str(args.thread),
                                  '-m',
                                  str(args.minimum_length),
                                  '--nextseq-trim=' + str(args.nextseq_trim),
                                  '--overlap',
                                  str(args.overlap),
                                  '-l',
                                  str(args.insert),
                                  '-o',
                                  out_fq2,
                                  args.fq]
    cutadapt.logger.info('%s' % (' '.join(cmd)))
    res = subprocess.run(cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    with open(args.outdir + '/cutadapt.log', 'wb') as fh:
        fh.write(res.stdout)

    format_stat(args.outdir + '/cutadapt.log', args.sample)

    step.clean_up()    


def get_opts_cutadapt(parser, sub_program):
    parser.add_argument('--adapter_fasta', help='addtional adapter fasta file')
    parser.add_argument('--minimum_length',dest='minimum_length',help='minimum_length', default=20)
    parser.add_argument('--nextseq_trim', help='nextseq_trim', default=20)
    parser.add_argument('--overlap',help='minimum overlap length', default=10)
    parser.add_argument('--insert', help="read2 insert length", default=150)
    if sub_program:
        parser.add_argument('--fq', help='fq file', required=True)
        parser.add_argument('--gzip', help="output gzipped fastq", action='store_true')
        parser = s_common(parser)
    return parser





"""
Integration tests
"""
import unittest
import os
import subprocess
import shutil
import glob



class Tests(unittest.TestCase):

    def setUp(self):
        self.test_dir = "/SGRNJ/Database/script/tests/multi_tests/"

    def _run(self, assay):
        os.chdir(self.test_dir + assay)
        print('running ' + assay)
        subprocess.check_call('sh run_shell.sh', shell=True)
        subprocess.check_call('sh sjm.sh', shell=True)
        if os.path.exists("test1"):
            shutil.rmtree("test1")
        subprocess.check_call('sh ./shell/test1.sh', shell=True)

    #@unittest.skip("pass")
    def test_rna(self):
        self._run('rna')
        assert os.path.exists("test1/06.analysis/stat.txt")

    def test_vdj(self):
        self._run('vdj')
        stat_file = glob.glob("test1/*.count_vdj/stat.txt")[0]
        assert os.path.exists(stat_file)


if __name__ == '__main__':
    unittest.main()



__STEPS__ = [
    'sample',
    'barcode',
    'cutadapt',
    'STAR',
    "featureCounts",
    "count",
    'analysis']
__ASSAY__ = 'rna'

# m: memory 
# x: thread
RESOURCE = {
    'sample': {'m':1, 'x':1},
    'barcode': {'m':5, 'x':1},
    'cutadapt': {'m':5, 'x':1},
    'STAR': {'m':30, 'x':1},
}



'''
vdj mapping
'''

import os
import pandas as pd
import matplotlib as mpl
import pysam
mpl.use('Agg')
from matplotlib import pyplot as plt

from celescope.tools.Step import Step, s_common
from celescope.vdj.__init__ import CHAINS
import celescope.tools.utils as utils


@utils.add_log
def summary(input_file, alignments, type, outdir, sample, assay, debug, not_consensus):
    chains = CHAINS[type]

    '''
    # out files
    UMI_unfiltered_file = f'{outdir}/{sample}_UMI_unfiltered.tsv'
    UMI_filtered1_file = f'{outdir}/{sample}_UMI_filtered1.tsv'
    UMI_filtered2_file = f'{outdir}/{sample}_UMI_filtered2.tsv'
    '''

    UMI_count_unfiltered_file = f'{outdir}/{sample}_UMI_count_unfiltered.tsv'
    UMI_count_filtered1_file = f'{outdir}/{sample}_UMI_count_filtered1.tsv'

    stat_prefix = 'UMIs'
    if not_consensus:
        stat_prefix = 'Reads'

    # read input_file
    with pysam.FastxFile(input_file) as fh:
        index = 0
        read_row_list = []
        for entry in fh:
            attr = entry.name.split("_")
            barcode = attr[0]
            umi = attr[1]
            dic = {"readId": index, "barcode": barcode, "UMI": umi}
            read_row_list.append(dic)
            index += 1
        df_read = pd.DataFrame(read_row_list, columns=["readId", "barcode", "UMI"])
        summary.logger.info(f"{input_file} to dataframe done.")
        total_read = df_read.shape[0]

    # init row list
    mapping_summary_row_list = []

    # mapped
    alignment = pd.read_csv(alignments, sep="\t")
    alignment.readId = alignment.readId.astype(int)
    align_read = alignment.shape[0]
    df_read.readId = df_read.readId.astype(int)
    df_align = pd.merge(df_read, alignment, on="readId", how="right")

    mapping_summary_row_list.append({
        "item": f"{stat_prefix} Mapped to Any VDJ Gene",
        "count": align_read,
        "total_count": total_read,
    })

    # CDR3
    df_CDR3 = df_align[~pd.isnull(df_align["aaSeqCDR3"])]
    align_read_with_CDR3 = df_CDR3.shape[0]
    mapping_summary_row_list.append({
        "item": f"{stat_prefix} with CDR3",
        "count": align_read_with_CDR3,
        "total_count": total_read,
    })

    # correct CDR3
    df_correct_CDR3 = df_CDR3[~(df_CDR3["aaSeqCDR3"].str.contains(r"\*"))]
    align_read_with_correct_CDR3 = df_correct_CDR3.shape[0]
    mapping_summary_row_list.append({
        "item": f"{stat_prefix} with Correct CDR3",
        "count": align_read_with_correct_CDR3,
        "total_count": total_read,
    })

    # VDJ
    df_VJ = df_correct_CDR3[
        (~pd.isnull(df_correct_CDR3['bestVGene'])) &
        (~pd.isnull(df_correct_CDR3['bestJGene']))
    ]
    df_VJ = df_VJ[df_VJ.bestVGene.str[:3] == df_VJ.bestJGene.str[:3]]
    df_VJ["chain"] = df_VJ.bestVGene.str[:3]
    df_VJ["VJ_pair"] = df_VJ["bestVGene"] + "_" + df_VJ["bestJGene"]
    Reads_Mapped_Confidently_to_VJ_Gene = df_VJ.shape[0]
    mapping_summary_row_list.append({
        "item": f"{stat_prefix} Mapped Confidently to VJ Gene",
        "count": Reads_Mapped_Confidently_to_VJ_Gene,
        "total_count": total_read
    })

    # chain
    for chain in chains:
        df_chain = df_VJ[df_VJ.chain == chain]
        Reads_Mapped_to_chain = df_chain.shape[0]
        mapping_summary_row_list.append({
            "item": f"{stat_prefix} Mapped to {chain}",
            "count": Reads_Mapped_to_chain,
            "total_count": total_read,
        })

    # unique UMI
    df_UMI = df_VJ.drop_duplicates(subset=["barcode", "UMI"], keep="first")

    # filter1: keep top 1 in each combinations
    groupby_elements = [
        'barcode',
        'chain',
        'bestVGene',
        'bestJGene',
        'aaSeqCDR3',
        'nSeqCDR3',
    ]
    df_UMI_count = df_UMI.groupby(
        groupby_elements, as_index=False).agg({"UMI": "count"})
    df_UMI_count = df_UMI_count.sort_values("UMI", ascending=False)
    # out unfiltered
    df_UMI_count.to_csv(UMI_count_unfiltered_file, sep="\t", index=False)

    df_UMI_count_filter1 = df_UMI_count.groupby(
        ["barcode", "chain"], as_index=False).head(1)
    # out filtered1
    df_UMI_count_filter1.to_csv(
        UMI_count_filtered1_file,
        sep="\t",
        index=False)

    '''
    #TODO
    if debug:
        unique_UMI = df_UMI.shape[0]
        mapping_summary_row_list.append({
            "item": "UMI unique count",
            "count": unique_UMI,
            "total_count": align_read_with_correct_CDR3,
        })
        UMI_after_Contamination_Filtering = df_UMI_count.filter1.UMI.sum()
        mapping_summary_row_list.append({
            "item": "UMI after Contamination Filtering",
            "count": UMI_after_Contamination_Filtering,
            "total_count": unique_UMI,
        })
    '''

    # stat file
    df = pd.DataFrame(
        mapping_summary_row_list,
        columns=[
            "item",
            "count",
            "total_count"])
    stat_file = f'{outdir}/stat.txt'
    utils.gen_stat(df, stat_file)


@utils.add_log 
def mixcr(outdir, sample, input_file, thread, species):
    report = f"{outdir}/{sample}_align.txt"
    not_align_fq = f"{outdir}/not_align.fq"
    read2_vdjca = f"{outdir}/read2.vdjca"
    alignments = f"{outdir}/{sample}_alignments.txt"

    cmd = f"""
mixcr align \
--force-overwrite \
--species {species} \
-t {thread} \
--not-aligned-R1 {not_align_fq} \
--report {report} \
-OallowPartialAlignments=true \
-OvParameters.geneFeatureToAlign=VTranscriptWithP \
{input_file} \
{read2_vdjca}
mixcr exportAlignments \
{read2_vdjca} {alignments} \
-readIds --force-overwrite -vGene -dGene -jGene -cGene \
-nFeature CDR3 -aaFeature CDR3\n"""
    mixcr.logger.info(cmd)
    os.system(cmd)
    return alignments


@utils.add_log
def mapping_vdj(args):

    step_name = f"{args.type}_mapping_vdj"
    step = Step(args, step_name)

    sample = args.sample
    outdir = args.outdir
    fq = args.fq
    receptor_type = args.type
    debug = args.debug
    assay = args.assay
    thread = int(args.thread)
    not_consensus = args.not_consensus
    species = args.species

    input_file = fq
    alignments = mixcr(outdir, sample, input_file, thread, species)

    # summary
    summary(input_file, alignments, receptor_type, outdir, sample, assay, debug, not_consensus)

    step.clean_up()



def get_opts_mapping_vdj(parser, sub_program):
    parser.add_argument("--type", help='TCR or BCR', required=True)
    parser.add_argument('--species', choices=['hs', 'mmu'], help='human or mouse', default='hs')
    parser.add_argument("--not_consensus", action='store_true', help="input fastq is not consensus")
    if sub_program:
        parser = s_common(parser)
        parser.add_argument("--fq", required=True)




from celescope.tools.report import reporter
from celescope.tools.utils import *
import matplotlib.pyplot as plt
import os
import pandas as pd
import logging
import numpy as np
import argparse
import glob
from collections import defaultdict
import matplotlib as mpl
mpl.use('Agg')

class Count_cite():

    def __init__(
        self,
        sample,
        outdir,
        assay,
        read_count_file,
        match_dir,
        ):
        self.sample = sample
        self.outdir = outdir
        self.assay = assay
        self.read_count_file = read_count_file
        self.match_dir = match_dir
        self.match_barcode, self.cell_total = read_barcode_file(match_dir)
        self.df_read_count = pd.read_csv(read_count_file, sep="\t", index_col=0)
        self.tsne_file = glob.glob(f'{match_dir}/*analysis/*tsne_coord.tsv')[0]
        
        if not os.path.exists(outdir):
            os.system('mkdir -p %s' % outdir)

        # out
        self.mtx = f'{outdir}/{sample}_citeseq.mtx.gz'

    @add_log
    def run(self):
        stats = pd.Series()
        outdir = self.outdir
        sample = self.sample
        UMI_tag_file = f'{outdir}/{sample}_umi_tag.tsv'
        mapped_read = self.df_read_count['read_count'].sum()

        # in cell
        df_read_count_in_cell = self.df_read_count[self.df_read_count.index.isin(self.match_barcode)]
        mapped_read_in_cell = int(df_read_count_in_cell['read_count'].sum())
        stats = stats.append(pd.Series(
            format_stat(mapped_read_in_cell, mapped_read),
            index=['Mapped Reads in Cells']
        ))

        # UMI
        df_UMI_in_cell = df_read_count_in_cell.reset_index().groupby([
            'barcode', 'barcode_name']).agg({'UMI': 'count'})
        df_UMI_in_cell = df_UMI_in_cell.reset_index()
        df_UMI_in_cell = df_UMI_in_cell.pivot(
            index='barcode', columns='barcode_name', values='UMI')
        df_cell = pd.DataFrame(index=self.match_barcode)
        df_UMI_cell = pd.merge(
            df_cell,
            df_UMI_in_cell,
            how="left",
            left_index=True,
            right_index=True)

        # fillna
        df_UMI_cell.fillna(0, inplace=True)
        df_UMI_cell = df_UMI_cell.astype(int)
        df_UMI_cell_out = df_UMI_cell.T
        df_UMI_cell_out.to_csv(self.mtx, sep='\t', compression='gzip')

        # UMI
        UMIs = df_UMI_cell.apply(sum, axis=1)
        median = round(np.median(UMIs), 2)
        mean = round(np.mean(UMIs), 2)
        stats = stats.append(pd.Series(
            str(median),
            index=['Median UMI per Cell']
        ))

        stats = stats.append(pd.Series(
            str(mean),
            index=['Mean UMI per Cell']
        ))

        self.stats = stats

    def report(self):

        self.stat_file = f'{self.outdir}/stat.txt'
        self.stats.to_csv(self.stat_file, sep=':', header=False)
        t = reporter(
        name='count_cite', 
        assay=self.assay, 
        sample=self.sample,
        stat_file=self.stat_file, 
        outdir=self.outdir + '/..')
        t.get_report()


import argparse
from .Count_cite import Count_cite

def get_opts_count_cite(parser, sub_program):
    parser.add_argument("--match_dir", help="matched scRNA-Seq CeleScope directory path", required=True)
    if sub_program:
        parser.add_argument('--outdir', help='output dir', required=True)
        parser.add_argument('--sample', help='sample name', required=True)
        parser.add_argument('--assay', help='assay', required=True)
        parser.add_argument("--read_count_file", help="tag read count file")


def count_cite(args):

    count_cite_object = Count_cite(
        args.sample,
        args.outdir,
        args.assay,
        args.read_count_file,
        args.match_dir,
    )
    count_cite_object.run()
    count_cite_object.report()


import unittest
import os
import pandas as pd
from collections import namedtuple
from celescope.tools.Step import Step
from .consensus import *


class Tests(unittest.TestCase):
    def setUp(self):
        pass

    def test_Step(self):
        Args = namedtuple("Args", 'sample outdir assay debug thread')
        args = Args(
            sample='test1',
            outdir='01.target_metrics',
            assay='snp',
            debug=False,
            thread=4,
        )
        step_name = "target_metrics"
        step = Step(args, step_name)

        step.add_metric(name="number value", value=1234)
        step.add_metric(name="fraction", fraction=0.123456)
        step.add_metric(name="str value", value="some str")
        step.add_metric(name="value with total", value=1234, total=3456)

        step.clean_up()

    def test_get_read_length(self):
        read_list = [['AAAA','FFFF'],['TTT','FFF'],['CCC','FFF'],['GGGGGGG','FFFFFFF']]
        assert get_read_length(read_list, 0.5) == 4

    def test_dumb_consensus(self):
        read_list = [('AAAA','FFFF'),('TTT','FF;'),('CCC','FFF'),('GGGGGGG','FFFFFFF')]
        consensus_seq, consensus_qual, ambiguous_base_n, con_len = dumb_consensus(read_list, 0.5)
        print(consensus_qual)
        assert consensus_seq == 'NNNA'


if __name__ == '__main__':
    unittest.main()


import os
import sys
import json
import functools
import gzip
import random
from collections import defaultdict
from itertools import groupby
import numpy as np
import pandas as pd
import subprocess
from scipy.io import mmwrite
from scipy.sparse import csr_matrix, coo_matrix
import pysam
from celescope.tools.utils import add_log, format_number, glob_genomeDir, gene_convert, s_common, add_mem
from celescope.tools.cellranger3.cell_calling_3 import cell_calling_3
from celescope.tools.__init__ import MATRIX_FILE_NAME, FEATURE_FILE_NAME, BARCODE_FILE_NAME
from celescope.tools.cellranger3 import get_plot_elements
from celescope.tools.Step import Step

toolsdir = os.path.dirname(__file__)
random.seed(0)
np.random.seed(0)


@add_log
def report_prepare(count_file, downsample_file, step_obj):

    df0 = pd.read_table(downsample_file, header=0)
    step_obj.add_data_item(percentile=df0['percent'].tolist())
    step_obj.add_data_item(MedianGeneNum=df0['median_geneNum'].tolist())
    step_obj.add_data_item(Saturation=df0['saturation'].tolist())
    step_obj.add_data_item(chart=get_plot_elements.plot_barcode_rank(count_file))
    step_obj.add_data_item(umi_summary=True)

def hd(x, y):
    return len([i for i in range(len(x)) if x[i] != y[i]])


def correct_umi(fh1, barcode, gene_umi_dict, percent=0.1):
    res_dict = defaultdict()

    for geneID in gene_umi_dict:
        _dict = gene_umi_dict[geneID]
        umi_arr = sorted(
            _dict.keys(), key=lambda x: (_dict[x], x), reverse=True)
        while True:
            # break when only one barcode or umi_low/umi_high great than 0.1
            if len(umi_arr) == 1:
                break
            umi_low = umi_arr.pop()

            for u in umi_arr:
                if float(_dict[umi_low]) / _dict[u] > percent:
                    break
                if hd(umi_low, u) == 1:
                    _dict[u] += _dict[umi_low]
                    del (_dict[umi_low])
                    break
        res_dict[geneID] = _dict
    return res_dict


@add_log
def bam2table(bam, detail_file):
    # 提取bam中相同barcode的reads，统计比对到基因的reads信息
    #
    samfile = pysam.AlignmentFile(bam, "rb")
    with gzip.open(detail_file, 'wt') as fh1:
        fh1.write('\t'.join(['Barcode', 'geneID', 'UMI', 'count']) + '\n')

        # pysam.libcalignedsegment.AlignedSegment
        # AAACAGGCCAGCGTTAACACGACC_CCTAACGT_A00129:340:HHH72DSXX:2:1353:23276:30843
        # 获取read的barcode
        def keyfunc(x): return x.query_name.split('_', 1)[0]

        for _, g in groupby(samfile, keyfunc):
            gene_umi_dict = defaultdict(lambda: defaultdict(int))
            for seg in g:
                (barcode, umi) = seg.query_name.split('_')[:2]
                if not seg.has_tag('XT'):
                    continue
                geneID = seg.get_tag('XT')
                gene_umi_dict[geneID][umi] += 1
            res_dict = correct_umi(fh1, barcode, gene_umi_dict)

            # output
            for geneID in res_dict:
                for umi in res_dict[geneID]:
                    fh1.write('%s\t%s\t%s\t%s\n' % (barcode, geneID, umi,
                                                    res_dict[geneID][umi]))
    samfile.close()


@add_log
def cell_calling(cell_calling_method, force_cell_num, expected_cell_num, all_matrix_10X_dir, df_sum, outdir, sample):
    if (force_cell_num is not None) and (force_cell_num != 'None'):
        cell_bc, UMI_threshold = force_cell(force_cell_num, df_sum)
    elif cell_calling_method == 'auto':
        cell_bc, UMI_threshold = auto_cell(df_sum, expected_cell_num)
    elif cell_calling_method == 'cellranger3':
        cell_bc, UMI_threshold = cellranger3_cell(all_matrix_10X_dir, expected_cell_num, df_sum)
    elif cell_calling_method == 'inflection':
        _cell_bc, UMI_threshold = auto_cell(df_sum, expected_cell_num)
        cell_bc, UMI_threshold = inflection_cell(outdir, sample, all_matrix_10X_dir, df_sum, UMI_threshold)
    cell_calling.logger.info(f'UMI_threshold: {UMI_threshold}')
    return cell_bc, UMI_threshold


@add_log
def force_cell(force_cell_num, df_sum):
    force_cell_num = int(force_cell_num)
    cell_range = int(force_cell_num * 0.1)
    cell_low = force_cell_num - cell_range
    cell_high = force_cell_num + cell_range

    df_barcode_count = df_sum.groupby(
        ['UMI']).size().reset_index(
        name='barcode_counts')
    sorted_df = df_barcode_count.sort_values("UMI", ascending=False)
    sorted_df["barcode_cumsum"] = sorted_df["barcode_counts"].cumsum()
    for i in range(sorted_df.shape[0]):
        if sorted_df.iloc[i, :]["barcode_cumsum"] >= cell_low:
            index_low = i - 1
            break
    for i in range(sorted_df.shape[0]):
        if sorted_df.iloc[i, :]["barcode_cumsum"] >= cell_high:
            index_high = i
            break
    df_sub = sorted_df.iloc[index_low:index_high + 1, :]
    threshold = df_sub.iloc[np.argmax(
        np.diff(df_sub["barcode_cumsum"])), :]["UMI"]
    cell_bc = get_cell_bc(df_sum, threshold, col='UMI')

    return cell_bc, threshold


def find_threshold(df_sum, idx):
    return int(df_sum.iloc[idx - 1, df_sum.columns == 'UMI'])


@add_log
def auto_cell(df_sum, expected_cell_num):
    col = "UMI"
    idx = int(expected_cell_num * 0.01)
    barcode_number = df_sum.shape[0]
    idx = int(min(barcode_number, idx))
    if idx == 0:
        sys.exit("cell number equals zero!")
    # calculate read counts threshold
    threshold = int(find_threshold(df_sum, idx) * 0.1)
    threshold = max(1, threshold)
    cell_bc = get_cell_bc(df_sum, threshold)

    return cell_bc, threshold


@add_log
def cellranger3_cell(all_matrix_10X_dir, expected_cell_num, df_sum):
    cell_bc, initial_cell_num = cell_calling_3(all_matrix_10X_dir, expected_cell_num)
    threshold = find_threshold(df_sum, initial_cell_num)
    return cell_bc, threshold


@add_log
def inflection_cell(outdir, sample, all_matrix_10X_dir, df_sum, threshold):
    app = f'{toolsdir}/rescue.R'
    cmd = (
        f'Rscript {app} '
        f'--matrix_dir {all_matrix_10X_dir} '
        f'--outdir {outdir} '
        f'--sample {sample} '
        f'--threshold {threshold}'
    )
    inflection_cell.logger.info(cmd)
    subprocess.check_call(cmd, shell=True)
    out_file = f'{outdir}/{sample}_rescue.tsv'
    df = pd.read_csv(out_file, sep='\t')
    inflection = int(df.loc[:,'inflection'])
    threshold = inflection
    cell_bc = get_cell_bc(df_sum, threshold)

    return cell_bc, threshold


@add_log
def get_df_sum(df, col='UMI'):
    def num_gt2(x):
        return pd.Series.sum(x[x > 1])

    df_sum = df.groupby('Barcode').agg({
        'count': ['sum', num_gt2],
        'UMI': 'count',
        'geneID': 'nunique'
    })
    df_sum.columns = ['readcount', 'UMI2', 'UMI', 'geneID']
    df_sum = df_sum.sort_values(col, ascending=False) 
    return df_sum

def get_cell_bc(df_sum, threshold, col='UMI'):
    return list(df_sum[df_sum[col] >= threshold].index)

@add_log
def plot_barcode_UMI(df_sum, threshold, expected_cell_num, cell_num, outdir, sample, cell_calling_method, col='UMI'):
    out_plot = f'{outdir}/{sample}_barcode_UMI_plot.pdf'
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.plot(df_sum['UMI'])
    plt.hlines(threshold, 0, cell_num, linestyle='dashed')
    plt.vlines(cell_num, 0, threshold, linestyle='dashed')
    plt.title('cell_calling_method: %s, expected_cell_num: %s\n %s threshold: %s, cell num: %s' %
              (cell_calling_method, expected_cell_num, col, threshold, cell_num))
    plt.loglog()
    plt.savefig(out_plot)


def get_cell_stats(df_sum, cell_bc, marked_counts_file):
    df_sum.loc[:, 'mark'] = 'UB'
    df_sum.loc[df_sum.index.isin(cell_bc), 'mark'] = 'CB'
    df_sum.to_csv(marked_counts_file, sep='\t')
    CB_describe = df_sum.loc[df_sum['mark'] == 'CB', :].describe()

    return CB_describe


def write_matrix_10X(table, id_name, matrix_10X_dir):
    id = table.index.to_series()
    name = id.apply(lambda x: id_name[x])
    genes = pd.concat([id, name], axis=1)
    genes.columns = ['gene_id', 'gene_name']

    #write
    table.columns.to_series().to_csv(
        f'{matrix_10X_dir}/barcodes.tsv', index=False, sep='\t')
    genes.to_csv(
        f'{matrix_10X_dir}/genes.tsv', index=False, header=False, sep='\t')
    mmwrite(f'{matrix_10X_dir}/matrix', csr_matrix(table))
    return id, name


@add_log
def matrix_10X(df, outdir, sample, gtf_file, dir_name='matrix_10X', cell_bc=None):
    matrix_10X_dir = f"{outdir}/{sample}_{dir_name}/"
    if not os.path.exists(matrix_10X_dir):
        os.mkdir(matrix_10X_dir)
    id_name = gene_convert(gtf_file)

    if cell_bc is not None:
        df = df.loc[df['Barcode'].isin(cell_bc), :]
    
    df_UMI = df.groupby(['geneID','Barcode']).agg({'UMI':'count'})
    mtx= coo_matrix((df_UMI.UMI, (df_UMI.index.labels[0], df_UMI.index.labels[1])))
    id = df_UMI.index.levels[0].to_series()
    # add gene symbol
    name = id.apply(lambda x: id_name[x])
    genes = pd.concat([id, name], axis=1)
    genes.columns = ['gene_id', 'gene_name']

    barcodes = df_UMI.index.levels[1].to_series()
    genes.to_csv(f'{matrix_10X_dir}/{FEATURE_FILE_NAME}', index=False, sep='\t', header=False)
    barcodes.to_csv(f'{matrix_10X_dir}/{BARCODE_FILE_NAME}', index=False, sep='\t')
    mmwrite(f'{matrix_10X_dir}/{MATRIX_FILE_NAME}', mtx)
    return matrix_10X_dir


@add_log
def expression_matrix(df, cell_bc, outdir, sample, gtf_file):

    id_name = gene_convert(gtf_file)

    df.loc[:, 'mark'] = 'UB'
    df.loc[df['Barcode'].isin(cell_bc), 'mark'] = 'CB'
    CB_total_Genes = df.loc[df['mark'] == 'CB', 'geneID'].nunique()
    CB_reads_count = df.loc[df['mark'] == 'CB', 'count'].sum()
    reads_mapped_to_transcriptome = df['count'].sum()

    table = df.loc[df['mark'] == 'CB', :].pivot_table(
        index='geneID', columns='Barcode', values='UMI',
        aggfunc=len).fillna(0).astype(int)

    # convert id to name; write table matrix
    matrix_table_file = f"{outdir}/{sample}_matrix.tsv.gz"
    id = table.index.to_series()
    name = id.apply(lambda x: id_name[x])
    table.index = name
    table.index.name = ""
    table.to_csv(
        matrix_table_file,
        sep="\t",
        compression='gzip')
    return(CB_total_Genes, CB_reads_count, reads_mapped_to_transcriptome)


def get_summary(df, sample, Saturation, CB_describe, CB_total_Genes,
                CB_reads_count, reads_mapped_to_transcriptome,
                stat_file, outdir):

    # total read
    json_file = outdir + '.data.json'
    fh = open(json_file)
    data = json.load(fh)
    str_number = data['barcode_summary'][1][1].split("(")[0]
    valid_read_number = int(str_number.replace(",", ""))

    summary = pd.Series([0, 0, 0, 0, 0, 0, 0],
                        index=[
                            'Estimated Number of Cells',
                            'Fraction Reads in Cells',
                            'Mean Reads per Cell',
                            'Median UMI per Cell',
                            'Total Genes',
                            'Median Genes per Cell',
                            'Saturation',
    ])

    # 细胞数
    summary['Estimated Number of Cells'] = int(
        CB_describe.loc['count', 'readcount'])
    summary['Fraction Reads in Cells'] = '%.2f%%' % (float(
        CB_reads_count) / reads_mapped_to_transcriptome * 100)
    summary['Mean Reads per Cell'] = int(
        valid_read_number /
        summary['Estimated Number of Cells'])
    summary['Median UMI per Cell'] = int(CB_describe.loc['50%', 'UMI'])
    summary['Total Genes'] = int(CB_total_Genes)
    summary['Median Genes per Cell'] = int(CB_describe.loc['50%', 'geneID'])
    summary['Saturation'] = '%.2f%%' % (Saturation)
    # 测序饱和度，认定为细胞中的reads中UMI>2的reads比例
    need_format = [
        'Estimated Number of Cells',
        'Mean Reads per Cell',
        'Median UMI per Cell',
        'Total Genes',
        'Median Genes per Cell']
    for item in need_format:
        summary[item] = format_number(summary[item])
    summary.to_csv(stat_file, header=False, sep=':')


@add_log
def sub_sample(fraction, df_cell, cell_bc, cell_read_index):
    '''
    cell_bc - cell barcode
    cell_read_index - df_cell repeat index
    '''
    cell_read = df_cell['count'].sum()
    frac_n_read = int(cell_read * fraction)
    subsample_read_index = cell_read_index[:frac_n_read]
    index_dedup, counts = np.unique(subsample_read_index, return_counts=True)
    n_count_once = np.sum(counts == 1)
    # total = UMI
    umi_total = len(index_dedup)
    umi_saturation = round((1 - n_count_once / umi_total) * 100, 2)
    read_total = frac_n_read
    read_saturation = round((1 - n_count_once / read_total) * 100, 2)

    # gene median
    df_cell_subsample = df_cell.loc[index_dedup,]
    geneNum_median = float(df_cell_subsample.groupby('Barcode').agg({'geneID': 'nunique'}).median())

    return umi_saturation, read_saturation, geneNum_median


@add_log
def downsample(df, cell_bc, downsample_file):
    """saturation and median gene
    return fraction=1 saturation
    """
    df_cell = df.loc[df["Barcode"].isin(cell_bc), :]
    cell_read_index = np.array(df_cell.index.repeat(df_cell['count']))
    np.random.shuffle(cell_read_index)

    format_str = "%.2f\t%.2f\t%.2f\n"
    res_dict = {
        "fraction": [],
        "umi_saturation": [],
        "read_saturation": [],
        "median_gene": []
    }
    with open(downsample_file, 'w') as fh:
        fh.write('percent\tmedian_geneNum\tsaturation\n')
        fh.write(format_str % (0, 0, 0))
        for fraction in np.arange(0.1, 1.1, 0.1):
            umi_saturation, read_saturation, geneNum_median = sub_sample(fraction, df_cell, cell_bc, cell_read_index)
            fh.write(format_str % (fraction, geneNum_median, umi_saturation))
            format_float = lambda x: round(x / 100, 4)
            res_dict["fraction"].append(round(fraction, 1))
            res_dict["umi_saturation"].append(format_float(umi_saturation))
            res_dict["read_saturation"].append(format_float(read_saturation))
            res_dict["median_gene"].append(geneNum_median)

    return umi_saturation, res_dict


@add_log
def count(args):

    step_name = "count"
    step = Step(args, step_name)

    # args
    outdir = args.outdir
    sample = args.sample
    assay = args.assay
    force_cell_num = args.force_cell_num
    cell_calling_method = args.cell_calling_method
    expected_cell_num = int(args.expected_cell_num)

    # check
    if args.genomeDir and args.genomeDir != "None":
        _refFlat, gtf_file = glob_genomeDir(args.genomeDir)
    else:
        gtf_file = args.gtf

    # umi纠错，输出Barcode geneID  UMI     count为表头的表格
    count_detail_file = outdir + '/' + sample + '_count_detail.txt.gz'
    bam2table(args.bam, count_detail_file)

    df = pd.read_table(count_detail_file, header=0)

    # export all matrix
    dir_name = 'all_matrix'
    all_matrix_10X_dir = matrix_10X(df, outdir, sample, gtf_file, dir_name=dir_name)

    # df_sum
    df_sum = get_df_sum(df)

    # call cells
    cell_bc, threshold = cell_calling(cell_calling_method, force_cell_num, expected_cell_num, all_matrix_10X_dir, df_sum, outdir, sample)

    # plot
    cell_num = len(cell_bc)
    plot_barcode_UMI(df_sum, threshold, expected_cell_num, cell_num, outdir, sample, cell_calling_method, col='UMI')
    
    # get cell stats
    marked_counts_file = outdir + '/' + sample + '_counts.txt'
    CB_describe = get_cell_stats(df_sum, cell_bc, marked_counts_file)

    # export cell matrix
    matrix_10X(df, outdir, sample, gtf_file, dir_name='matrix_10X', cell_bc=cell_bc)
    (CB_total_Genes, CB_reads_count, reads_mapped_to_transcriptome) = expression_matrix(
            df, cell_bc, outdir, sample, gtf_file)

    # downsampling
    cell_bc = set(cell_bc)
    downsample_file = args.outdir + '/' + args.sample + '_downsample.txt'
    Saturation, res_dict = downsample(df, cell_bc, downsample_file)

    # summary
    stat_file = outdir + '/stat.txt'
    get_summary(df, sample, Saturation, CB_describe, CB_total_Genes,
                CB_reads_count, reads_mapped_to_transcriptome, stat_file,
                outdir + '/../')

    report_prepare(marked_counts_file, downsample_file, step)

    step.add_content_item('metric', downsample_summary=res_dict)
    step.clean_up()




def get_opts_count(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--bam', required=True)
        parser.add_argument('--force_cell_num', help='force cell number', default=None)
    parser.add_argument('--genomeDir', help='genome directory')
    parser.add_argument('--gtf', help='gtf file path')
    parser.add_argument('--expected_cell_num', help='expected cell number', default=3000)
    parser.add_argument('--cell_calling_method', help='cell calling methods', 
        choices=['auto', 'cellranger3', 'inflection',], default='auto')



import os
import sys
import json
import numpy as np
import pandas as pd
import celescope.tools.utils as utils
from celescope.tools.Step import Step


class AnalysisMixin():

    def __init__(self, args):
        if hasattr(args, "match_dir") and args.match_dir:
            self.match_dir = args.match_dir
        else:
            self.match_dir = args.outdir + "/../" # use self

    @staticmethod
    def get_cluster_tsne(colname, tsne_df, show_colname=True):
        """
        tSNE_1	tSNE_2	cluster Gene_Counts
        return data list
        """

        sum_df = tsne_df.groupby([colname]).agg("count").iloc[:, 0]
        percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
        res = []
        for cluster in sorted(tsne_df[colname].unique()):
            sub_df = tsne_df[tsne_df[colname] == cluster]
            if show_colname:
                name = f"{colname} {cluster}({percent_df[cluster]}%)"
            else:
                name = f"{cluster}({percent_df[cluster]}%)"
            tSNE_1 = list(sub_df.tSNE_1)
            tSNE_2 = list(sub_df.tSNE_2)
            res.append({"name": name, "tSNE_1": tSNE_1, "tSNE_2": tSNE_2})
        return res

    @staticmethod
    def get_table_dict(marker_table):
        table_dict = Step.get_table(
            title='Marker Genes by Cluster',
            table_id='marker_gene_table',
            df_table=marker_table,
        )
        return table_dict

    def get_marker_table(self):
        """
        return html code
        """
        marker_df = self.marker_df.loc[:,
            ["cluster", "gene", "avg_logFC", "pct.1", "pct.2", "p_val_adj"]
        ]
        marker_df["cluster"] = marker_df["cluster"].apply(lambda x: f"cluster {x}")

        return marker_df

    def get_gene_tsne(self):
        """
        return data dic
        """
        tsne_df = self.tsne_df
        tSNE_1 = list(tsne_df.tSNE_1)
        tSNE_2 = list(tsne_df.tSNE_2)
        Gene_Counts = list(tsne_df.Gene_Counts)
        res = {"tSNE_1": tSNE_1, "tSNE_2": tSNE_2, "Gene_Counts": Gene_Counts}
        return res

    def read_match_dir(self):
        match_dict = utils.parse_match_dir(self.match_dir)
        tsne_df_file = match_dict['tsne_coord']
        self.tsne_df = pd.read_csv(tsne_df_file, sep="\t")
        self.tsne_df.rename(columns={"Unnamed: 0": "barcode"}, inplace=True)
        marker_df_file = match_dict['markers']
        self.marker_df = pd.read_csv(marker_df_file, sep="\t")

    def run_analysis(self):
        self.read_match_dir()

        self.cluster_tsne = self.get_cluster_tsne(colname='cluster', tsne_df=self.tsne_df)
        self.gene_tsne = self.get_gene_tsne()
        marker_table = self.get_marker_table()
        self.table_dict = self.get_table_dict(marker_table)
